/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: grounder.cpp
 *
 *  CONTENTS:  Part of the BACK END, PSGRND.
 *             Implementation of the class 'Grounder'
 *
 *  DATE:     June 4, 2004
 ***************************************************************************/

#include "back_end.h"

/*
int Grounder::evaluate(Expression *expression)
DESCRIPTION
	Run-time evaluatior of postfix expressions using the stack 'stk'.
PARAMETERS
	expression    - Expression to evaluate
RETURNS
	TRUE  - Expression is evaluated correctly. If it contains relational
	         operands or other conditions, all of them are true. Also all
	         assignments (if there are any) are done successfully, i.e.
	         the assigned value was in in the range for the variable
	         being assigned.
	FALSE - If at least one condition for returning TRUE value did not hold.
	         The expression could not be evaluated correctly if there was
	         division by zero or there was weight function call with
	         undefined arguments. In such cases, the evaluator tries to
	         report an error (but the reporting in not done in every case:
	         see 'groundAtomList' method).
*/
int Grounder::evaluate(Expression *expression)
{
	int op1, op2;
	Expression *exp;
	DPnode *dp;

	// NOTE: if some operation has two operands, the second operand
	//  is the one that will be popped first !

	for(exp=expression ; exp ; exp=exp->next)
		switch(exp->token)
		{
		case EXPR_STRING:
		case EXPR_NUM:
			// At this moment string constant is
			// represented by the corresponding number
			stk.push(exp->num);
			break;
		case EXPR_VAR:
			// 'valIndex' field in the variable node contains the
			// index of the variable's value. The index is taken
			// in the array of arguments for the data type of
			//  the variable
			stk.push(exp->var->type->arguments[exp->var->valIndex]);
			break;
		case EXPR_EXPR:  // Sub-expression
			// Evaluate the sub-expression.
			if(!evaluate(exp->exp)) return 0;

			//Sub-expressions can be used only in the argument
			// lists. Since all the numbers corresponding to valid
			// arguments happen to be in the declaration sequence,
			// the result of this expression should also be a valid
			// number from the data sequence.
			//Otherwise, the argument is automatically out of range.
			//This check is essential; without it, the numerical
			// result of an expression may be confused with the
			// number corresponding to a string constant.
			if(!strConst.isNumber(stk.getElement(0))) return 0;
			break;

		case EXPR_PLUS: // Add two top elements in the stack
			stk.push(stk.pop()+stk.pop());
			break;
		case EXPR_MINUS:// Subtract the top from the element right below
			op2=stk.pop();
			stk.push(stk.pop()-op2);
			break;
		case EXPR_MUL:  // Multiply two top elements
			stk.push(stk.pop()*stk.pop());
			break;
		case EXPR_DIV:  // Divide the element below the top by the top
			op2=stk.pop();
			if(op2==0)  // Check for division by zero
			{
				  // Report an error and return FALSE
				reportError(GERR_DIVBYZERO, expression);
				return 0;
			}
			stk.push(stk.pop()/op2);
			break;
		case EXPR_MOD:  // Very similar to division
			op2=stk.pop();
			if(op2==0)
			{
				reportError(GERR_DIVBYZERO, expression);
				return 0;
			}
			stk.push(stk.pop()%op2);
			break;

		case EXPR_MIN:  // Find the minimum between the two top elements
			op2=stk.pop();
			op1=stk.pop();
			stk.push((op1>=op2)?op2:op1);
			break;
		case EXPR_MAX: // Find the maximum between the two top elements
			op2=stk.pop();
			op1=stk.pop();
			stk.push((op1>=op2)?op1:op2);
			break;

		case EXPR_ABS:    // Absolute value of the top
			op1=stk.pop();
			stk.push((op1<0)?(-op1):op1);
			break;
		case EXPR_UMINUS: // Apply the unary minus to the top
			stk.push(-stk.pop());
			break;

		case EXPR_WEIGHT:  // Evaluate weight function
			// The expression for a weight function call is built
			//  the way that the arguments precede the expression
			//  containing a weight function. Thus, at this moment
			//  the N top elements in the stack should contain
			//  evaluated argument list
			// (where N == 'exp->wFun->arity)
			  // Search the corresponding vector in the
			  //  arguments of the weight function
			op1=stk.vectorSearch(exp->wFun->arity,
				exp->wFun->arguments, exp->wFun->valNum);
			  // If the vector was not found, the weight function
			  // is not defined on the given set of arguments.
			  // Report an error in this case:
			if(op1<0)
			{
				reportError(GERR_NOWEIGHT, expression);
				return 0;
			}
			  // If found the vector, the weight function value
			  //  should be stored in the array 'results' at
			  // the index, equal to the INDEX OF THE VECTOR,
			  // which is "op1 / arityOfWeightFunction".
			stk.push(exp->wFun->results[op1/exp->wFun->arity]);
			break;

		case EXPR_DPOSITIVE:  // Positive data predicate condition
			dp=exp->atom->dp;
			  // If the argument list of data predicate cannot be
			  //  evaluated (some expressions give values that
			  //  couldn't occur in the argument list) or
			  //  the data predicate doesn't have the given
			  //  set of arguments in the range, then return FALSE,
			  // otherwise, proceed.
			if( !evaluate(exp->atom->arg) ||  stk.vectorSearch(
			     dp->arity, dp->arguments, dp->valNum) < 0 )
			        return 0;
			break;
		case EXPR_DNEGATIVE: // Negative data predicate condition
			dp=exp->atom->dp;
			  // It should be true if and only if the same
			  //  but positive data predicate condition is false.
			if( evaluate(exp->atom->arg) && stk.vectorSearch(
			      dp->arity, dp->arguments, dp->valNum) >= 0 )
			        return 0;
			break;

		case EXPR_ASSIGN:  // Assignment operation
			// NOTE: Value that is being assigned to a variable
			//  always stays in the stack.
			  // Since the assignment is done to some variable that
			  //  has type, the assigned number should not be the
			  //  one that couldn't occur in the data sequence:
			if(!strConst.isNumber(stk.getElement(0))) return 0;
			 // Do the type checking and assign the index to
			 //  the 'VarNode::valIndex':
			if((exp->var->valIndex = stk.scalarSearch(
			     stk.getElement(0), exp->var->type->arguments,
			     exp->var->type->valNum))<0)
				return 0;
			break;

		default:  // This ought to be a comparison operation:
			op2=stk.pop();  // Right operand
			op1=stk.pop();  // Left operand
			  // If we are dealing with the special case:
			if(EXPR_CMP_IS_SPECIAL(exp->num))
			{
				// If one of the expressions is a string and the
				//  other is a number, need intervention to make
				//  the comparison work correctly:
				  // If left operand always has to be a number
				  // or it is number because it is in the range:
				if( exp->num == EXPR_CMP_LEFTNUM ||
				    strConst.isNumber(op1) )
				{  // The left operand is a number.
					  // If the right operand doesn't have
					  //  to be a number and it is not
					  // a number: (=> 'op2' is a string)
					  //  MAKE 'op1' LESS THAN 'op2':
					if( exp->num != EXPR_CMP_RIGHTNUM &&
					   !strConst.isNumber(op2) )
					{op1=-1; op2=1;} // 'op2' is a string
				}
				else
				{  // The left operand is a string.
					  // If the right operand(op2)
					  //  is a number,
					  //  MAKE op2 LESS THAN op1;
					if( exp->num == EXPR_CMP_RIGHTNUM ||
					    strConst.isNumber(op2) )
					{op2=-1; op1=1;}
				}
			}

			  // Perform the relational operation:
			switch(exp->token)
			{
			case EXPR_LESS:       if(op1 >= op2) return 0; break;
			case EXPR_NOTLESS:    if(op1 <  op2) return 0; break;
			case EXPR_GREATER:    if(op1 <= op2) return 0; break;
			case EXPR_NOTGREATER: if(op1 >  op2) return 0; break;
			case EXPR_EQUAL:      if(op1 != op2) return 0; break;
			case EXPR_NOTEQUAL:   if(op1 == op2) return 0; break;
			default:
				fprintf(stderr, "CRITICAL ERROR in "
				    "Grounder::evaluate\n"
				    "wrong type of the expression node\n\n");
				exit(1);
			}
		}

	return 1;
}


/*
int Grounder::setNextValue(Quantifier *quant)
DESCRIPTION
	This method is used for quantifier instantiation.
	Every quantifier has a current value associated with it. This method
	sets the next value to the given quantifier
PARAMETERS
	quant    -  Quantifier which value will be set
RETURNS
	TRUE     -  Success: the value has been set
	FALSE    -  Failure: the current value is the last value of the
	                      quantifer; there are no next values.
*/
int Grounder::setNextValue(Quantifier *quant)
{
	VarNode *var;
	DPnode *dp;
	Expression *exp;
	int i, dpValue;
	int errFlg;

	switch(quant->token)
	{
	// If the quantifier is empty, it is considered to have only one
	//  legal value: when "Quantifier::dpValIndex == 0"
	case QUANT_EMPTY:
		if(++(quant->dpValIndex)>0) return 0;
		return 1;

	// Variable quantifiers has as many values as the quantified variable
	//  Moreover, the current value is indicated by "VarNode::valIndex",
	//  unlike the other quantifiers which use "Quantifier::dpValIndex"
	case QUANT_VAR:
		var=quant->var;
		var->valIndex++;  // Go to the next value.
			// Check whether this is a valid index:
		if(var->valIndex >= var->type->valNum) return 0;
		else return 1;

	// Data predicate quantifier is the most difficult case. It may quantify
	//  several variables and also its arguments can be known variables,
	//  constants, or arithmetic expressions (with all known variables at
	//  that point). Therefore, in addition to taking the next value we also
	//  need to assign the values to the quantified variables and also make
	//  sure that the values of the arguments other than quantified
	//  variables match the corresponding arguments of the data predicate
	case QUANT_DP:
		  // Starting from the next value of the quantifier go over all
		  //  values until all conditions are satisfied or there are no
		  // more values of the data predicate. All conditions are
		  //  satisfied if type checking holds for values of all
		  //  quantified variables and the all the other arguments
		  //  match the corresponding argument of the data predicate.
		for(dp=quant->dp, quant->dpValIndex+=dp->arity ;
			quant->dpValIndex < dp->valNum*dp->arity ;
			quant->dpValIndex+=dp->arity)
		{
			  // Go over the argument list and at the same time over
			  //  the current vector of values of the arguments
			  // As soon as some condition is not satisfied,
			  // 'errFlg' will be set to 1 and the loop breaks.
			for(errFlg=0, i=quant->dpValIndex, exp=quant->arg ;
			    exp && !errFlg ; exp=exp->next, i++)
			{
				  // Get the value of the current argument:
				dpValue=dp->arguments[i];

				  // Next, do the action depending on the type
				  //  of the argument.
				switch(exp->token)
				{
				// If quantified variable, try to assign the
				//  value by finding the index of the value
				//  in the variable's data type:
				case EXPR_VAR_QUANT:
					exp->var->valIndex = stk.scalarSearch(
					  dpValue, exp->var->type->arguments,
					  exp->var->type->valNum);

					if(exp->var->valIndex < 0) errFlg=1;
					break;

				// If this is a number or string, check whether
				//  it is equal to the corresponding argument:
				case EXPR_NUM:
					if(exp->num != dpValue) errFlg=1;
					break;

				// If this is a known variable, check whether
				//  its value is equal to the argument:
				case EXPR_VAR:
					if( exp->var->type->arguments[
					   exp->var->valIndex] != dpValue)
						errFlg=1;
					break;

				// If this is an arithmetic expression, then it
				//  can be equal only if the argument is not a
				//  string constant, expression can be evaluated
				//  and the result is equal to the argument:
				case EXPR_EXPR:
					if(!strConst.isNumber(dpValue) ||
					   !evaluate(exp->exp) ||
					   stk.pop()!=dpValue ) errFlg=1;

					// Clear the stack after the operation:
					stk.clear();
					break;
				default:
					fprintf(stderr, "CRITICAL ERROR in the "
					  "method Grounder::setNextValue\n"
					  "Incorrectly constructed quantifier"
					  "\n\n");
					exit(1);
				}
			}

			// If we went through the whole argument list without
			// violating any condition, then the current value of
			// the data predicate is exactly what we need.
			if(!errFlg) return 1;
		}

		// We can be here only if we passed the last set of arguments
		//  for the data predicate. Need to return failure.
		return 0;
	}

	return 0;  // Prevent compilers from generating warnings
}


/*
void Grounder::resetValue(Quantifier *quant)
DESCRIPTION
	This method is used for quantifier instantiation.
	Every quantifier has a current value associated with it. This method
	assigns the value "-1" to the quantifier, i.e. such value that
	subsequent call of the method 'setNextValue' will set the first
	value of the quantifier
PARAMETERS
	quant    -  Quantifier to reset
*/
void Grounder::resetValue(Quantifier *quant)
{
	switch(quant->token)
	{
	//For data predicate, 'setNextValue' increments 'Quantifier::dpValIndex'
	// by data predicate's arity => set it to '-arity' so that the next
	// value will be zero.
	case QUANT_DP:
		quant->dpValIndex=-quant->dp->arity;
		break;
	// Set the correspoding index to '-1' so the
	//  'setNextValue' will make it zero
	case QUANT_EMPTY:
		quant->dpValIndex=-1;
		break;
	case QUANT_VAR:
		quant->var->valIndex=-1;
		break;
	}
}


/*
int Grounder::instantiate(Quantifier *list, Quantifier * &current)
DESCRIPTION
	Important method of the grounder. It instantiates a list of quantifers,
	i.e. assigns a set of values to the quantified variables

PARAMETERS
	list     -  a list of quantifiers to instantiate
	current  -  A reference to a pointer to the current
	             quantifier (see below)
RETURNS
	TRUE     -  Success: A set of values is assigned
	FALSE    -  Failure: No more values to assign

HOW IT WORKS
	1) Let 'listOfQuantifiers' will be the list that we need to instantiate.
	    Declare a pointer to 'Quantifier' node (e.g. 'cur')
	    and initialize it to NULL

	2) instantiate(listOfQuantifiers, cur): First such call will assign the
	    first set of values to the variables; each subsequent call will
	    assign the next set of values to the variables. This works while
	    method returns TRUE. Once it returns FALSE, there are no
	    more values to assign. (NOTE: 'cur' should not be modified between
	    the calls of the method).
	    NOTE: If the 'listOfQuantifiers' is empty or there are no quantified
	    variables (i.e. there is only one empty quantifier) the method
	    returns TRUE after the first call and FALSE after the second.
*/
int Grounder::instantiate(Quantifier *list, Quantifier * &current)
{
	static Quantifier dummyNode;

	    // If this is a first call with list==NULL (i.e. 'current' is also
	    //  NULL) make 'current' point to the dummy node and return TRUE.
	    // During the subsequent call with an empty list, it will set
	    //  'current' back to NULL and return FALSE.
	if(!list)
	{
		if(!current) { current=&dummyNode; return 1; }
		else { current=0; return 0;}
	}

	    // If this is the first call with the given quantifier list,
	    //  set the 'current' to the first quantifer and reset its value
	if(!current)
	{
		current=list;
		resetValue(list);
	}

	while(1)  // Main loop
	{
		    // Assign the next value to 'current' quantifier.
		    // If this cannot be done (the value of 'current' is the
		    //  last value, go back along the list of quantifiers
		    //  until we find the one that could be assigned the
		    //  next value.
		while(!setNextValue(current))
		{
			current=current->prev; // Go back
			   // If none of the quantifier can be assigned the
			   // next value, there are no more sets of values
			   // to be assigned =>  return FALSE
			if(!current) return 0;
		}

		stk.clear(); // Need to clear the stack sometimes

		    // Evaluate the condition list of the current quantifier
		    //  for the next value assigned on the previous step.
		    // If conditions are satisfied, go to the next quantifier
		    //  in the list.
		if(evaluate(current->conditions))
		{
			    // If this was a last quantifier, then we have
			    //  a set of values assigned => return TRUE
			if(!current->next) return 1;
			    // Otherwise go to the next quantifier
			    //  and reset its value
			current=current->next;
			resetValue(current);
		}
	}
}


/*
int Grounder::run(RuleNode *theory)
DESCRIPTION
	Main method of the grounder. Grounds the whole theory.

PARAMETERS
	theory - list of rules to ground

RETURNS
	0    -  Everything is OK.
	1    -  There were errors during grounding
*/
int Grounder::run(RuleNode *theory)
{
	Quantifier *cur;
	int groundRuleCount;	// Variable to count ground rules.

	stk.clear(); // Clear the evaluation stack at the beginning

		// Loop over the theory: "FOR EVERY RULE..."
	for(groundRuleCount=0 ; theory ; theory=theory->next)
	{
		lineNum=theory->lineNum;

			// Instantiate global quantifiers for the given rule:
		for(cur=0;instantiate(theory->quant, cur);)
		{
				// Start rule in the ground theory.
			out.startRule(theory->horn, lineNum, ++groundRuleCount);

				// Ground the simple atoms in the body:
			groundAtomList(theory->body);
				// Ground the cardinality and
				//  weight atoms in the body:
			groundMetaAtomList(theory->cardBody,   0);
			groundMetaAtomList(theory->weightBody, 1);


				// Start grounding head:
			out.switchToHead();

				// Ground simple atoms in the head:
			groundAtomList(theory->head);
				// Ground cardinality and
				//  weight atoms in the head:
			groundMetaAtomList(theory->cardHead,   0);
			groundMetaAtomList(theory->weightHead, 1);

				// Finish grounding rule. If the conflict
				//  was found, stop the grounding.
			if(!out.endRule()) return (errorCount!=0);
		}

			// Save the ground rule number of the last rule:
		theory->maxGroundRule=groundRuleCount;
	}

	stk.clear();  // Clear the stack before returning.
	return (errorCount!=0);
}


/*
int groundMetaAtomList(MetaAtom *mAtom, int isWeight)
DESCRIPTION
	Ground a list of cardinality or weight atoms.

PARAMETERS
	mAtom    -  First atom in the list of meta atoms
	isWeight -  True if this is a list of weight atoms,
	             false if the list of cardinality atoms
*/
void Grounder::groundMetaAtomList(MetaAtom *mAtom, int isWeight)
{
	int min, max;
	for(; mAtom ; mAtom=mAtom->next)
	{
			// If there was an error evaluating the expression
			//  (most probably division by zero):
		if(!evaluate(mAtom->min) || !evaluate(mAtom->max))
		{
			stk.clear();  // Clear the stack after the evaluator
			// ERROR RECOVERY: Make the atom to be always false:
			min=1; max=0;
		}
		else // Otherwise, get the evaluated limits
		{
			max=stk.pop();
			min=stk.pop();
		}

			// Output beginning of the meta-atom
		out.startMeta(min, max, isWeight);
		groundAtomList(mAtom->list);   // Ground meta list
	}
}


/*
void Grounder::groundAtomList(Atom *list)
DESCRIPTION
	The following method grounds a list of program atoms
	The list can be in the body, head, or inside cardinality
	or weight atom.
PARAMETERS
	list   -  List of atoms to be grounded
*/
void Grounder::groundAtomList(Atom *list)
{
	int weight=0;
	Quantifier *cur;

	for(; list ; list=list->next) // For every atom in the list:
	{
			// Instantiate local quantifiers:
		for(cur=0 ; instantiate(list->quant, cur) ;)
		{
				// Clear the evaluation stack
				//  after the instantiation:
			stk.clear();

				// If we are dealing with a weight assignment,
				//  process the weight expression:
			if(list->weight)
			{
				  // If there is an error during evaluation of
				  //  the weigh expression (division by zero or
				  //  undefined weight function for the given
				  //  arguments), ERROR RECOVERY: go to the
				  //  next atom (thus, this one is automatically
				  //  considered to be false)
				lastErrorCode=GERR_NOERROR; //Reset the variable
				if(!evaluate(list->weight))
				{
				      // If the evaluator couldn't report the
					  // error, (happens when an expression
					  // in an argument of a weight function
					  // evaluates to a number that couldn't
					  // occur in the declaration sequence),
					  // this method needs to report the
					  //  error itself:
				    if(lastErrorCode == GERR_NOERROR)
				        reportError(GERR_NOWEIGHT,list->weight);

				    continue;
				}

				weight=stk.pop();   // Get the weight

				  // Weight cannot be negative. Use the same
				  // ERROR RECOVERY as in previous case
				if(weight <= 0)
				{
				  reportError(GERR_BADWEIGHT,list->weight,list);
				  // ERROR RECOVERY: this atom is assumed
				  //  to be false
				  continue;
				}
			}

				// Add the ground atom to the theory:
			out.addAtom(list, weight);
		}

		stk.clear(); // Clear the stack again
	}

		// Finish output of the atom list:
	out.endAtomList();
}


/*
void Grounder::reportError(int flag, Expression *exp, Atom *atom)
DESCRIPTION
	This method reports grounding errors
PARAMETERS
	flag    -  Error code. The codes are defined in the header file
	exp     -  Condition list where the error occurred. This is used
	            to print the current values of program variables
	atom    -  Not used at this moment
*/
void Grounder::reportError(int flag, Expression *exp, Atom *atom)
{
	static TraverseVariables goVars;
	VarNode *var;
	int wasBefore;

	lastErrorCode=flag;  // Save the last error code

		// This makes grounder to suppress error messages when
		//  there are too many of them.
	if(errorCount < MAX_GROUNDING_ERRORS)
	{
		errorCount++;  // Increase the error count

			// Print the beginning of the message:
		fprintf(errorStream, "GROUDING ERROR: rule in line %d: ",
			lineNum);

			// Print the end of the message:
		switch(flag)
		{
		case GERR_DIVBYZERO:
			fprintf(errorStream, "division by zero ");
			break;
		case GERR_NOWEIGHT:
			fprintf(errorStream, "weight function "
				"cannot be evaluated ");
			break;
		case GERR_BADWEIGHT:
			fprintf(errorStream, "weight is less than "
				"or equal to zero ");
			break;
		default:
			fprintf(stderr, "CRITICAL ERROR in "
				"Grounder::reportError\n");
			exit(1);
		}

		// PRINT THE VALUES OF THE PROGRAM VARIABLES:
			// Set 'tmpFlag' field of all variables in
			//  the expression to zero:
		for(goVars.start("E", exp);(var=goVars.next());) var->tmpFlag=0;

			// Print the variables:
		for(goVars.start("E", exp),wasBefore=0; (var=goVars.next());)
				// If this is not a temporary variable,
				//   print its name and contents:
			if(var->symbolno > 0)
			{
				     // If this is the first time, print the
				     //  messages; otherwise, print a separator
				if(wasBefore) fprintf(errorStream, ", ");
				else
					fprintf(errorStream, "for the following"
					  " values of the program variables: ");
				wasBefore=1;

				    // Print the name of the variable:
				fprintf(errorStream, "%s=",
					symbt.getName(var->symbolno));

				    // If the value of 'var->valIndex' is valid,
				    //  print the value of the variable;
				    //  otherwise, print that it is not assigned
				if(var->valIndex >= 0 &&
				   var->valIndex < var->type->valNum)
					strConst.printConstant(errorStream,
					   var->type->arguments[var->valIndex]);
				else
					fprintf(errorStream, "<NOT ASSIGNED>");
			}

		fprintf(errorStream, "\n");

			// If there were too many error messages, inform the
			//  user that the subsequent ones will be suppressed:
		if(errorCount == MAX_GROUNDING_ERRORS)
			fprintf(errorStream, "TOO MANY ERRORS during grounding:"
				" all similar messages will be suppressed\n");
	}
}
