/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  MODIFIED BY: Artur Mikitiuk
 *
 *  FILE NAME: main.cpp
 *
 *  PROGRAM NAME:  GROUNDER FOR THE ANSWER SET PROGRAMMING SYSTEM
 *                 PROPOSITIONAL SCHEMATA WITH PSEUDOBOOLEAN CONSTRAINTS
 *  RELEASE DATE & VERSION:                                                */
#define VERSION "Version: psgrnd 7-Jul-2005"
/***************************************************************************/

#include <stdio.h>
#include "rules.h"
#include "back_end.h"

// Default output file name (different for Windows and Unix):
#ifdef _WIN32
#define DEFAULT_OUTPUT "out.txt"
#else
#define DEFAULT_OUTPUT "out.aspps"
#endif

// 2 in the 32-nd power. Needed to convert clock count.
#define POW_2_32	(double)0x100000000

// Class timer to count how much time it took to run the grounder.
//  Different for Windows and Unix:
#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
class Timer  // TIMER for Windows
{
	LARGE_INTEGER Begin; // Starting time:
	double Freq;         // Timer frequency
public:
		// This method (re)starts the timer:
	void restart() { QueryPerformanceCounter(&Begin); }

		// This method reads how much time elapsed from
		//  the last call of the 'restart' method.
	double read()
	{
		LARGE_INTEGER End;
		double Time;
		QueryPerformanceCounter(&End); // Read the timer
			// Save convert the ending time to double:
		Time=(double)End.LowPart+(double)End.HighPart*POW_2_32;
			// Subtract the starting time:
		Time-=(double)Begin.LowPart+(double)Begin.HighPart*POW_2_32;
			// Adjust according to frequency:
		return(Time/Freq);
	}

		// Constructor: Intialize the frequency and restart the timer:
	Timer()
	{
		QueryPerformanceFrequency(&Begin); // Read the frequency
			// Save the frequency:
		Freq=(double)Begin.LowPart+((double)Begin.HighPart)*POW_2_32;
		restart();
	}
};
#else
#include <time.h>
class Timer  // TIMER for Unix:
{
	clock_t begin;  // Starting time
public:
		// This method (re)starts the timer:
	void restart() {begin=clock();}

		// This method reads how much time elapsed from
		//  the last call of the 'restart' method.
	double read()
	{
		clock_t end;
		double time;
		end=clock(); // Read the timer
			// Calculate and return the elapsed time:
		time=((double)(end-begin))/((double)CLOCKS_PER_SEC);
		return time;
	}

		// Constructor: read the timer:
	Timer() { begin=clock();}
};
#endif

// The only global variables in the program:
//  Objects of the classes the program consists of:
SymbolTable st;
StringConstants strConst(st);
LexicalAnalyzer la(st);
Rules rules(st, strConst, la);
GroundRules groundRules(st, strConst);
Grounder grounder(st, groundRules, strConst);

extern int yyparse();  // Parser constructed by YACC.
// Auxiliary method for the 'main' function. It analyzes the string, tells
//  whether it is constant assignment, and if it is, loads the constant:
int loadConstant(SymbolTable &st, char *string);

// Message that is given by the grounder if it is executed
//  without any command line parameters:
char usage[]=
"Usage: psgrnd [-d dataFileList] -r ruleFile [-c constantList] [output] [flags]\n"
"output: [-o outputForSolver] [-m humanReadableOutput]\n"
"flags:  [-DM ] [-DL ] [-DP] [-DN] [-S] [-T] [-D] [-P] [-dimacs] [-pbs]\n"
"NOTE: flags -d,-r,-c may be dropped, but in this case they should be dropped\n"
"      altogether and the data files should always precede the rule file\n"
"NOTE: if \'-o ..\' is not specified, the default output file is \"%s\"\n"
"      if \'-o ..\' is given without file name, the Standard Output will be used\n"
"-d dataFile1 dataFile2 ... dataFileN : One or more data files\n"
"-c name1=value1 name2=value2 ... : Input constants from the command line\n"
"-DL : Disable one-step lookahead(performed by default after grounding)\n"
"-DP : Disable propagation (includes effect of the flags -DL and -DM)\n"
"-DM : Disable memory compaction. The rules that were eliminated during\n"
"      propagation and lookahead physically remain in place, so it is possible\n"
"      to trace how the ground theory was obtained\n"
"-DN : Disable the new output format, where atoms that don't occur in ground\n"
"      rules are assigned the number \'-1\'\n"
"-S : Enable scripting. This option requires having human-readable output\n"
"     specified. This allows to monitor the process of propagation and lookahead\n"
"     by printing all the atoms that were assigned some value to the\n"
"     human-readable output file\n"
"-D or -dimacs: Generate output in DIMACS format\n"
"-P or -pbs: Generate output in PBS format\n"
"-T : Print how much time it took to execute the grounder\n";

// Additional macros to be used by the function 'main'. These are created
//  to avoid writing the same code several times. All of them print
//  error messages for the command line parameters:
#define WRONG_COMMAND_LINE_PARAMETER(str) \
{fprintf(stderr, "ERROR: Wrong command line parameter: \'%s\'\n", str); \
errorCount++;}

#define CANNOT_OPEN_INPUT_FILE(str) \
{fprintf(stderr, "ERROR: Cannot open the input file: \"%s\"\n", str); \
errorCount++;}

#define CANNOT_OPEN_FILE_FOR_WRITING(str) \
{fprintf(stderr, "ERROR: Cannot open file \"%s\" for writing\n", str); \
errorCount++;}

#define OPTION_SHOULD_BE_FOLLOWED_BY_THE_FILENAME(opt) \
{fprintf(stderr, "ERROR: Option \'%s\' should be followed by " \
"the name of the output file\n", opt);  errorCount++;}

// Main function reads the command line parameters and then executes
//  the other parts of the program:
int main(int argc, char *argv[])
{
	Timer timer; // To time if necessary
		// Indicates whether the next expected argument
		//  is an option, data file, or constant (see below):
	int readingMode;

		// TRUE if new mode of command line parametes is used;
		//  FALSE otherwise:
	int newMode;
	int i, errorCount, returnValue, len;
	int outputFlag=0, isTimerEnabled=0, isScriptEnabled=0;
	FILE *mainOut=0, *humanOut=0, *mainOut2=0;
	char *mainOutName="";
	char *mainOut2Name="";
	char *ruleFile=0;


	timer.restart();  // Restart the timer.
	fprintf(stderr, "%s\n", VERSION); // Print the version of the program.
		// If there are no command line arguments,
		//  print the messages how to use the program and quit:
	if(argc<=1)
	{
		fprintf(stderr, usage, DEFAULT_OUTPUT);
		return 0;
	}

		// Check what mode of command line arguments is used: old or new
		// If either of "-r", "-d", or "-c" is used, then the old
		//  output format; otherwise, new output format:
	for(i=1, newMode=1;i<argc;i++)
		if(argv[i][0]=='-' &&
			(argv[i][1]=='r' || argv[i][1]=='d' || argv[i][1]=='c')
			 && argv[i][2]==0)
		{ newMode=0; break;}


		// Read the command line parameters:
		// Values for readingMode:
		//      0 - nothing
		//      1 - reading list of data files
		//      2 - reading list of constants
	for(i=1, readingMode=0, errorCount=0 ; i<argc &&
		errorCount <= MAX_CMDLINE_ERRORS ; i++)
	{
			// If the first element of the current parameter
			//  is '-', it resets reading mode back to reading
			//  the command line arguments:
		if(readingMode && argv[i][0] == '-') readingMode=0;

			// If we are currently reading the data files, try to
			//  open the file given in the parameter. If there is
			//  an error while opening, print the error message:
		if(readingMode==1 && la.addFile(argv[i]))
			CANNOT_OPEN_INPUT_FILE(argv[i])

			// If we are currently reading a list of constants,
			//  load the constant. If it fails, print the
			//  error message:
		if(readingMode==2 && !loadConstant(st, argv[i]))
		{
			fprintf(stderr, "ERROR: Wrong constant "
				"assignment: \"%s\"\n", argv[i]);
			errorCount++;
		}

			// If we are not reading command line arguments,
			//  skip everything that is below:
		if(readingMode) continue;

		if(argv[i][0]=='-' && argv[i][1]!=0)
		{
			if(argv[i][2]==0)
			{ // We are here if the command line option has exactly
			  //  one character after '-'.
				switch(argv[i][1])
				{
				case 'r':
				// Don't open the rule file right away:
				//  it should be opened only after all the data files:
					if(i+1>=argc || argv[i+1][0]=='-')
						OPTION_SHOULD_BE_FOLLOWED_BY_THE_FILENAME("r")
					else
						ruleFile=argv[++i];
					break;
				case 'd':
				// Check whether this option is followed by a
				//  filename (there should be at least one)
				//  and switch the mode to read a
				//  list of data files:
					if(i+1>=argc || argv[i+1][0]=='-')
						OPTION_SHOULD_BE_FOLLOWED_BY_THE_FILENAME("d")
					else
						readingMode=1;
					break;
				case 'c': readingMode=2; break;
				case 'D': outputFlag|=GROUT_DIMACS; break;
				case 'P': outputFlag|=GROUT_PBS; break;
				case 'T': isTimerEnabled=1; break;
				case 'S': isScriptEnabled=1; break;
				case 'o':
				// Make sure the option is followed by output
				//  file name and try to open the file:
					if(i+1>=argc || argv[i+1][0]=='-')
						mainOut=stdout;
					else
						if(!(mainOut=fopen((mainOutName=argv[++i]), "w")))
							CANNOT_OPEN_FILE_FOR_WRITING(argv[i])
					break;

				case 'm':
				// Make sure the option is followed by output
				//  file name and try to open the file:
					if(i+1>=argc || argv[i+1][0]=='-')
						OPTION_SHOULD_BE_FOLLOWED_BY_THE_FILENAME("m")
					else
						if(!(humanOut=fopen(argv[++i], "w")))
							CANNOT_OPEN_FILE_FOR_WRITING(argv[i])
					break;
				default:
					WRONG_COMMAND_LINE_PARAMETER(argv[i]);
				}
			}
			else
			  // We are here if the command line option
			  //  has more than one character after '-'
				if(argv[i][1]=='D' && argv[i][2]!=0 &&
				   argv[i][3]==0)
				{ // Read "-D?" command line arguments:
					switch(argv[i][2])
					{
					case 'M': groundRules.disableAllCompaction(); break;
					case 'P': groundRules.disableAnyPropagation(); break;
					case 'L': groundRules.disableLookahead(); break;
					case 'N': outputFlag|=GROUT_LEGACY; break;
					default:
						WRONG_COMMAND_LINE_PARAMETER(argv[i]);
					}
				}
				else if (!strcmp(argv[i]+1,"dimacs"))
				    outputFlag|=GROUT_DIMACS; 
				else if (!strcmp(argv[i]+1,"pbs"))
				    outputFlag|=GROUT_PBS; 
				else
					WRONG_COMMAND_LINE_PARAMETER(argv[i]);
		}
		else
		{ // We are here if a command line option expected but there
		  //  was no '-' followed by an option.
		  // This means an error in case of old format and a name of
		  //  the input file (can be data or rules) in case of
		  //  the new format:
			if(!newMode)
				WRONG_COMMAND_LINE_PARAMETER(argv[i])
			else
				if(!loadConstant(st, argv[i]))
				{
					ruleFile=argv[i];
					if(la.addFile(ruleFile))
						CANNOT_OPEN_INPUT_FILE(argv[i])
				}
		}
	}

	// NOTE: All the following steps are performed only if there were no
	//        errors in the command line parameters:

		// There has to be exactly one rule file specified:
	if(!errorCount && !ruleFile)
	{
		fprintf(stderr, "ERROR: NO RULE FILES (should be exactly one)\n");
		errorCount++;
	}

		// If this is old mode, then open the rule file (in the new
		//  mode it is already opened). If can not open the file,
		//  print the error message:
	if(!errorCount && !newMode && la.addFile(ruleFile))
	{
		CANNOT_OPEN_INPUT_FILE(ruleFile)
		errorCount++;
	}

		// If main output file was not specified, open the file with
		//  the default name. If it can not be done, print the message:
	if(!errorCount && !mainOut)
	{
		mainOutName=DEFAULT_OUTPUT;
		mainOut=fopen(mainOutName, "w");
		if(!mainOut)
		{
			CANNOT_OPEN_FILE_FOR_WRITING(DEFAULT_OUTPUT)
			errorCount++;
		}
	}

		// If the script is enabled, there has to be human-readable
		//  output file specified. If not, print the error message:
	if(!errorCount && isScriptEnabled)
	{
		if(humanOut) groundRules.enableScript(humanOut);
		else
		{
			fprintf(stderr, "ERROR: Cannot enable script if "
				"human output file is not specified\n");
			errorCount++;
		}
	}

	if (outputFlag&GROUT_PBS) {
	    if (mainOut==stdout) {
		fprintf(stderr, "ERROR: output in the PBS format "
			"cannot be sent to the standard output device\n");
		errorCount++;
	    } else {
		len=strlen(mainOutName);
		mainOut2Name=(char *)malloc(len+4);
		strcpy(mainOut2Name, mainOutName);
		strcpy(mainOut2Name+len, ".pb");
		mainOut2=fopen(mainOut2Name, "w");
		if (!mainOut2) {
		    CANNOT_OPEN_FILE_FOR_WRITING(mainOut2Name)
		    errorCount++;
		}
	    }
	}

	// PERFORM THE GROUNDING:
	if(!errorCount)
	{
	    returnValue=yyparse(); // Parse the input files:

		// For debugging purposes:
	    if(DEBUG_DONTFREE) rules.test();
	    if(DEBUG_PRINT_RULES_AFTER) rules.printTheory();

	    if(!returnValue) // If there were no errors during parsing:
	    {
		    // If there were some warnings, print the
		    //  newline to separate the messages during
		    //  parsing from the messages during
		    //  grounding:
		if(rules.getNumberOfWarnings())
		    fprintf(rules.getErrorStream(), "\n");

		    // Ground the rules. If there were errors during
		    //  grounding, set the flag GROUT_ERRORS:
		if(grounder.run(rules.getRules()))
		    returnValue=GROUT_ERRORS;

		    // Add the flags specifying output
		returnValue|=outputFlag;

		    // Execute the final stage of grounding. If
		    //  it is successful and the output was
		    //  created, inform the user about it:
		if(groundRules.executeFinalStage(returnValue,
		    rules.getProgramPredicateList(), mainOut, mainOut2,
		    humanOut, rules.getRules()) && mainOut!=stdout) {
		    if (outputFlag&GROUT_PBS)
			fprintf(stderr, "SUCCESS: Created files \"%s\" and "
			    "\"%s\"\n", mainOutName, mainOut2Name);
		    else
			fprintf(stderr, "SUCCESS: Created file \"%s\"\n"
			    ,mainOutName);
		    }
	    }

		// If the timer is enabled, print how much time
		//  it took to ground the theory:
	    if(isTimerEnabled)
		fprintf(stderr, "It all took us %.4f seconds\n", timer.read());
	}

	la.closeAllFiles(); // Close all the input files
		// Close all the output files:
	if(humanOut) fclose(humanOut);
	if(mainOut) fclose(mainOut);
	if(mainOut2) fclose(mainOut2);
	humanOut=mainOut=mainOut2=0;

	return 0; // Exit the program
}


/*
int loadConstant(SymbolTable &st, char *string)
DESCRIPTION
	The function checks the string for being assignment to a constant.
	If check is successful, it loads the constant to the symbol table.
PARAMETERS
	st      -  Instance of the symbol table to load the constant into
	string  -  String that may contain the constant assignment
RETURNS
	TRUE  - 'string' was in fact an assignment to a constant and
	         it has been loaded successfully
	FALSE - 'string' is not an assignment or there is an error in it.
*/
// Simple, but useful macros:
#define IS_LETTER(c) ((c)>='A' && (c) <='Z' || (c) >= 'a' && (c) <='z' \
                                            || (c) == '_')
#define IS_DIGIT(c) ((c)>='0' && (c)<='9')
int loadConstant(SymbolTable &st, char *string)
{
	int i, equality, symbolno;

		// Checking whether the first part
		//  of the string is an identifier:
	if(!IS_LETTER(string[0])) return 0; // First character has to be letter
		// The remaining characters can be either digit or letter
	for(i=0; IS_LETTER(string[i]) || IS_DIGIT(string[i]); i++);

		// The identfier should be followed by '='
	if(string[i] != '=') return 0;

		// The rest of the string should be a number =>
		//  all characters have to be digits.
	for(equality=i, i++; IS_DIGIT(string[i]); i++);

		// Check whether this number has at least one digit
		//  and is followed by the end of string:
	if(i==equality+1 || string[i]!=0) return 0;

		// At this point we made sure that we are dealing with correct
		//  constant assignment.
		// Put the end-of-string element to be able
		//   to read the identifier correctly:
	string[equality]=0;

		// Load the constant into the symbol table:
	symbolno=st.insertName(string);
	st.setType(symbolno, TYPE_NUM_CONST);
	st.getVal(symbolno)->number=atoi(string + equality +1);

		// Restore the string to its original state:
	string[equality]='=';

	return 1; // Indentifier has been loaded successfully.
}
