/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: ruleserr.cpp
 *
 *  CONTENTS:  Part of the FRONT END, PSGRND. This file contains all the
 *              methods from the class 'Rules' that print error messages.
 *             There are 2 parts:
 *             1) Error reporting and handling methods
 *             2) Metods for printing debug information
 *
 *  DATE:     March 30, 2004
 ***************************************************************************/

#include <stdarg.h>
#include "rules.h"


/****************************************************************************
*****************************************************************************
******    ERROR REPORTING AND HANDLING METHODS for the class 'Rules'
*****************************************************************************
****************************************************************************/


// Macro that prints the name of the given variables. For regular program
//  variables this is a name taken from the symbol table. For temporary
//  variables, the name is 'Tmp' and the temporary variable number
#define PRINT_VARIABLE(var) \
	{if((var)->symbolno>0) printf("%s",symbt.getName((var)->symbolno)); \
	 else printf("Tmp%d", -(var)->symbolno);}


/*
int Rules::semanticError(int code, ...)
DESCRIPTION
	The method that prints all semantic error messages generated
	in the action handlers or auxiliary methods. The arguments vary
	depending on the type of error. The description of errors and
	corresponding arguments to that method can be found
	in the file "struc.h".
PARAMETERS
	code   -  Semantic message code. Determnines the rest of the
	           parameters and behaviour of the method
RETURNS
	NOT USED
*/

// Read 2 or 3 numbers from the arguments in "..." to the variables
//  num1, num2, [and num3].
#define READ_TWO_NUMBERS {num1=va_arg(args, int); num2=va_arg(args, int);}
#define READ_THREE_NUMBERS {READ_TWO_NUMBERS num3=va_arg(args, int);}

// Make sure 'num' is a valid index in the array 'types'. Otherwise,
//  make 'num' to be the index of the last element
#define CHECK_NUMBER_FOR_ET(num) \
	{if(num>=ARRAY_SIZE(types)) num=ARRAY_SIZE(types)-1;}

// Do the previous thing for 'num1' and 'num2'.
#define CHECK_2_NUMBERS_FOR_ET \
	CHECK_NUMBER_FOR_ET(num1) CHECK_NUMBER_FOR_ET(num2)

// Set the variable given in the 'name' to the string that describes a type
//  of some entry in the symbol table given by its symbol number
#define DETERMINE_OBJECT_TYPE(symbolno, name) \
		switch(symbt.getType(symbolno)) \
		{ \
		case TYPE_PROGRAM_PRED: name="program predicate";break; \
		case TYPE_DATA_PRED: name="data predicate"; break; \
		case TYPE_WEIGHT_FUN:name="weight function";break; \
		default: name="UNKNOWN"; \
		}

// Self-explanatory:
#define NAME_META_LIMIT(isMaximum) ((isMaximum)?"maximum":"minimum")
#define NAME_META_TYPE(type) (((type)==ATOM_WEIGHT)?"weight atom":"cardinality atom")

// THE METHOD ITSELF
int Rules::semanticError(int code, ...)
{
        // Variables to handle arguments in "..."
    va_list args;
    va_start(args, code);

    static char ErrStr[1000]; // String for 'sprintf' for simple messages.

        // Array containing the names of the types of expessions
        // as they given in the expression handler node:
    static char *types[]=ET_TYPE_NAME_ARRAY;

        // The variable 'specialCase' indicates how
        //  the error message should be printed:
        //  a) 'specialCase==0': for simples error messages (by default)
        //    the body of message is expected to be copied into 'ErrStr',
        //    inside the main 'switch' opearatior and then at the end
        //    it will be printed with the word "WARNING" or "ERROR"
        //    and the current line number. Also either error count or
        //    warning will be increased accordingly
        //  b) 'specialCase==1': error message will be printed inside
        //     the 'switch', so don't need to print it after (mostly
        //     done for messages which don't need the line number or
        //     words "ERROR" or "WARNING". Still, need to increase
        //     error count or warning count.
        //  c) 'specialCase==2': Neither print nor increase the count:
        //     everything should be done inside the 'switch'.
    int specialCase=0;

        // Other variables necessary for printing the messages:
    PPnode *pp;
    int num1, num2, num3;
    ExpressionHandler *eh;
    int num, i, isWeight;
    WFnode *wn;
    DPnode *dp;
    VarNode *var;
    Value *val;
    char *name, *oldName;
    void *stuff;

    ErrStr[0]=0;  // The string will be empty.

        // MAIN SWITCH
        //  The switch for every possible value of 'code':
    switch(code)
    {
        // ERRORS:
    case SE_DIV_ZERO: //semanticError(SE_DIV_ZERO, Operation);
        sprintf(ErrStr, "division by zero");
        break;

    case SE_EXP_TYPES:
            // semanticError(SE_EXP_TYPES, int leftType, int rightType);
        READ_TWO_NUMBERS
        CHECK_2_NUMBERS_FOR_ET
        sprintf(ErrStr, "cannot combine expressions of the types: %s and %s",
            types[num1], types[num2]);
        break;

    case SE_EXP_UNKNOWN_ID:
            // semanticError(SE_EXP_UNKNOWN_ID, identifierName);
        name=va_arg(args, char *);
        sprintf(ErrStr, "found unknown identifier \'%s\'. Probably this is a "
            "constant that should be entered from the command line", name);
        break;

    case SE_RANGE_TYPES:
            // semanticError(SE_RANGE_TYPES, int lowerType, int upperType);
        READ_TWO_NUMBERS
        CHECK_2_NUMBERS_FOR_ET

        sprintf(ErrStr, "expressions in the data range both have to be NUMBER, "
                "but they are %s and %s", types[num1], types[num2]);
        break;

    case SE_RANGE_WRONG:
            // semanticError(SE_RANGE_WRONG, int LowerBound, int UpperBound);
        READ_TWO_NUMBERS
        sprintf(ErrStr, "wrong range \"%d..%d\": the lower bound can not "
                        "be greater than the upper bound", num1, num2);
        break;

    case SE_RANGE_BIG:
            // semanticError(SE_RANGE_BIG, int LowerBound, int UpperBound);
        READ_TWO_NUMBERS
        sprintf(ErrStr, "range \"%d..%d\" is too big:"
                " the program can\'t handle it", num1, num2);
        break;

    case SE_MANY_ARITIES:
            // semanticError(SE_MANY_ARITIES, int SymbolNumber,
            //                              int NewlyEncounteredArity);
        READ_TWO_NUMBERS
        DETERMINE_OBJECT_TYPE(num1, name);
        sprintf(ErrStr,"%s \'%s\' was previously declared with arity %d,"
            " but now it has arity %d", name, symbt.getName(num1),
            ((DPnode *)(symbt.getVal(num1)->ptr))->arity,
            num2);
        break;

    case SE_WEIGHT_EXP_TYPE:
        // semanticError(SE_WEIGHT_EXP_TYPE,int WeightFunctionSymbolNumber,
        //                            ExpressionHandler *TroublesomeExpression);
        sprintf(ErrStr, "wrong weight in the weight function declaration");
        break;

    case SE_PROBLEM_CONST:
        // semanticError(SE_PROBLEM_CONST, int MinimalNumberInTheProgram,
        //                                int NumberOfConstantsInTheProgram);
        READ_TWO_NUMBERS
        sprintf(ErrStr, "since the numbers used in the data sequence are"
                    " too small (as small as %d), the grounder cannot work "
                    "with %d string constants", num1, num2);
        break;

    case SE_TYPE_NOT_UNARY:
        //   Variable declaration:
        // semanticError(SE_TYPE_NOT_UNARY, int CausePredSymbNo, VarNode *vars)
        //   Program predicate declaration:
        // semanticError(SE_TYPE_NOT_UNARY, int CausePredSymbNo, 0)

        num1=va_arg(args, int);
        var=va_arg(args, VarNode *);

        sprintf(ErrStr, "data predicate \'%s\' cannot be used "
            "as a data type in declaration of the %s because it is not unary",
            symbt.getName(num1), (var)?"program variables":"data predicate");
        break;

    case SE_EXPR_DECLARATION:
        READ_TWO_NUMBERS
        CHECK_NUMBER_FOR_ET(num1)

        sprintf(ErrStr, "arithmetic expressions in the Data Sequence "
            "can contain only NUMBERS, but this one has the type %s",
                types[num1]);
        break;

    case SE_WEIGHT_IN_DECLARATION:
        // semanticError(SE_WEIGHT_IN_DECLARATION, symbolno);
        num1=va_arg(args, int);

        sprintf(ErrStr, "(weight call of \'%s\'): cannot call weight functions"
            " in data sequence", symbt.getName(num1));
        break;

    case SE_DATA_REPETITION:
        wn=va_arg(args, WFnode *);
        num=va_arg(args, int);
        val=va_arg(args, Value *);

        specialCase=2;
        isWeight=(symbt.getType(wn->symbolno) == TYPE_WEIGHT_FUN);
        if(isWeight) semanticErrorCount++;
        else warningCount++;

            // val==NULL means that no message should be printed.
        if(!val) break;

            // Print the data predicate or the weight function
        fprintf(errorStream, (isWeight)?"ERROR: Weight function %s(":
                "WARNING: Data predicate %s(", symbt.getName(wn->symbolno));

            // Print its parameters:
        for(i=0;i<wn->arity;i++)
        {
            if(i) fprintf(errorStream, ",");
            strConst.printConstant(errorStream, wn->arguments[num+i]);
        }

        fprintf(errorStream, ") is declared several times:\n");

        for(oldName=0;val;)
        {
            if(val->fileName!=oldName && lexan.isFileNameValid())
            {
                fprintf(errorStream, "%s: line(s): ", val->fileName);
                oldName=val->fileName;
            }

            for(;val && val->fileName==oldName; val=val->next)
                fprintf(errorStream, "%d, ",val->lineNum);

            fprintf(errorStream, "\n");
        }

        break;

    case SE_PP_BAD_TAIL:
        // semanticError(SE_PP_BAD_TAIL, PPnode *ProgramPredicate);
        pp=va_arg(args, PPnode *);
        sprintf(ErrStr, "data predicate \'%s\' (arity %d) cannot be a tail "
                "to the program predicate \'%s\' (arity %d)"
                " because they have different arities",
                symbt.getName(pp->tail->symbolno), pp->tail->arity,
                symbt.getName(pp->symbolno), pp->arity);
        break;

    case SE_MANY_PPVALS:
        // semanticError(SE_MANY_PPVALS, PPnode *ProgramPredicate)
        pp=va_arg(args, PPnode *);
        sprintf(ErrStr, "the number of possible combinations of argumens of the"
                        " program predicate \'%s\' is more than (2^32-1): "
                        "the grounder cannot work with such predicates",
                        symbt.getName(pp->symbolno));
        break;

    case SE_ARGUMENT_WEIGHT:
        // semanticError(SE_ARGUMENT_WEIGHT, symbolNumberOfThePredicate,
        //                                   argumentWithWeight);
        num1=va_arg(args, int);
        DETERMINE_OBJECT_TYPE(num1, name);

        sprintf(ErrStr, "argument of the %s \'%s\' contains weight function",
                        name, symbt.getName(num1));
        break;

    case SE_PARAM_NUM:
            // semanticError(SE_PARAM_NUM, symbolno, argList);
        num1=va_arg(args, int);
        DETERMINE_OBJECT_TYPE(num1, name);

        sprintf(ErrStr, "the number of arguments of the %s \'%s\'"
                        " differs from its arity",
                        name, symbt.getName(num1));
        break;

    case SE_PP_VAR_TYPE:
        // semanticError(SE_PP_VAR_TYPE, programPredicateNode,
        //                               argumentList, indexOfParameter);
        pp=va_arg(args, PPnode *);
        stuff=va_arg(args, void *);
        num1=va_arg(args, int);

        sprintf(ErrStr, "variable type mismatch in the argument #%d"
                        " of the program predicate \'%s\'",
                        num1+1, symbt.getName(pp->symbolno));
        break;

    case SE_BAD_HORN_RULE:
        // semanticError(SE_BAD_HORN_RULE, antecedent, consequent);
        sprintf(ErrStr, "wrong horn rule "
                        "(probably result of some previous error)");
        break;

    case SE_META_BAD_LIMIT:
        // semanticError(SE_META_BAD_LIMIT, typeOfAtom,
        //                                  isThisMaximum , expressionHandler);
        READ_TWO_NUMBERS
        eh=va_arg(args, ExpressionHandler *);
        num3=eh->type;
        CHECK_NUMBER_FOR_ET(num3)

        sprintf(ErrStr, "%s cannot be used as %s limit in a %s", types[num3],
                    NAME_META_LIMIT(num2), NAME_META_TYPE(num1));
        break;

    case SE_WEIGHTEXP_BAD:
        // semanticError(SE_WEIGHTEXP_BAD, simpleAtom, expressionHandler);
        stuff=va_arg(args, void *);
        eh=va_arg(args, ExpressionHandler *);
        num3=eh->type;
        CHECK_NUMBER_FOR_ET(num3)

        sprintf(ErrStr, "%s cannot be used as a weight in a weight assignment",
                        types[num3]);
        break;

    case SE_PATOM_WRONGTYPES:
        // semanticError(SE_PATOM_WRONGTYPES, HLeft, HRight);
        sprintf(ErrStr, "weight functions cannot be used in predefined atoms");
        break;

    case SE_HORN_ATOM_CONFLICT:
        // semanticError(SE_HORN_ATOM_CONFLICT, programPredicate);
        pp=va_arg(args, PPnode *);

        fprintf(errorStream, "ERROR: horn predicate \'%s\' appears in "
            "cardinality or weight atom in a horn rule\n",
            symbt.getName(pp->symbolno));

        specialCase=1;
        break;

    case SE_GLOBAL_LOCAL_MIX:
        // semanticError(SE_GLOBAL_LOCAL_MIX, temporaryAtom, variable);
        stuff=va_arg(args, void *);
        var=va_arg(args, VarNode *);

        sprintf(ErrStr, "variable \'%s\', which is global for the rule, "
            "appears in the local variable list",symbt.getName(var->symbolno));
        break;

    // WARNINGS:
    case SW_NUM_OUT_RANGE:
        // semanticError(SW_NUM_OUT_RANGE, number);
        num1=va_arg(args, int);

        sprintf(ErrStr, "the number %d cannot be an argument because it is"
            " less than any number in the data sequence", num1);
        break;

    case SW_PP_OUT_RANGE:
        // semanticError(SW_PP_OUT_RANGE, programPredicate,
        // argument, indexOfParameter);
        pp=va_arg(args, PPnode *);
        READ_TWO_NUMBERS

        num3=sprintf(ErrStr, "program atom with the predicate \'%s\' is always"
                             " false because its argument #%d is equal to ",
                             symbt.getName(pp->symbolno), num2+1);

        if(strConst.isNumber(num1))
            num3+=sprintf(ErrStr+num3, "%d", num1);
        else
            num3+=sprintf(ErrStr+num3, "\'%s\'",
                    strConst.getConstantFromNumber(num1));

        sprintf(ErrStr+num3, " which is not in the range"
                             " for the data type \'%s\'",
                             symbt.getName(pp->params[num2]->symbolno));
        break;

    case SW_META_NO_LIMITS:
        //semanticError(SW_META_NO_LIMITS,atomType,cList);
        num1=va_arg(args, int);
        sprintf(ErrStr, "%s has no limits", NAME_META_TYPE(num1));
        break;

    case SW_META_MAX_LESS_MIN:
        // semanticError(SW_META_MAX_LESS_MIN, atomType, minNum, maxNum);
        READ_THREE_NUMBERS

        sprintf(ErrStr, "%s is false because its maximum=%d is less than"
            " minimum=%d", NAME_META_TYPE(num1), num2, num3);
        break;

    case SW_META_TRUE:
        //  semanticError(SW_META_TRUE, atomType);
        num1=va_arg(args, int);
        sprintf(ErrStr, "%s is always true", NAME_META_TYPE(num1));
        break;

    case SW_META_NEGATIVE:
        // semanticError(SE_META_NEGATIVE, atomType, isMaximum, minNum);
        READ_THREE_NUMBERS

        if(num2)
            sprintf(ErrStr, "Maximim limit in %s is negative: the atom "
                    "will always be false", NAME_META_TYPE(num1));
        else
            sprintf(ErrStr, "Minimum limit in %s is negative: doesn't make"
                " sense to do this since this is equivalent to zero",
                NAME_META_TYPE(num1));
        break;

    case SW_LOCALVAR_UNUSED:
        // semanticError(SW_LOCALVAR_UNUSED, atom, var);
        stuff=va_arg(args, void *);
        var=va_arg(args, VarNode *);

        sprintf(ErrStr, "variable \'%s\', which appears in the local variable"
                    " list, doesn't appear in the argument list of the program "
                    "predicate and therefore ignored",
                    symbt.getName(var->symbolno));
        break;

    case SW_LOCALVAR_REPEAT:
        // semanticError(SW_LOCALVAR_REPEAT, atom, var);
        stuff=va_arg(args, void *);
        var=va_arg(args, VarNode *);

        sprintf(ErrStr, "variable \'%s\' is repeated in the local variable "
            "list; the repetition is ignored", symbt.getName(var->symbolno));
        break;

    case SW_USELESS_DATA_ATOM:
            // semanticError(SW_USELESS_DATA_ATOM, DataPredicate, Variable)
        dp=va_arg(args, DPnode *);
        var=va_arg(args, VarNode *);
        sprintf(ErrStr, "it doesn't make any sense to put atom \"%s(%s)\""
            " because it is always TRUE. Probably this is the old format.",
            symbt.getName(dp->symbolno), symbt.getName(var->symbolno));
        break;

    case SM_STOP_BEFORE_RULES:
        fprintf(errorStream, "PARSING IS STOPPED at the line %d because of "
            "errors in data and/or declarations\n", lexan.getLineNum(0));

        specialCase=1;
        break;

    case SM_INTERRUPTED:
        fprintf(errorStream, "Input files are parsed, but no output can be"
                    " generated because of errors\n");

        specialCase=1;
        break;

    case SM_SUPPRESS_REPEAT_MSG:
        fprintf(errorStream, "There were too many messages regarding "
           "repeated arguments. All subsequent messages will be suppressed\n");
        specialCase=1;
        break;

    default:
        sprintf(ErrStr, "unknown semantic error (code %d)", code);
    }

        // At this moment we need to write the error message and increase
        // error or warning count, except when we deal with special cases.
        // specialCase == 1 means that the message is already printed,
        //                  so nothing should be printed here.
        // specialCase == 2 same as (1), but in addition should not modify
        //                  error counters (they have already been modified.

    if(specialCase < 2)
    {
        if(SE_IS_ERROR(code)) semanticErrorCount++;
        else if(SE_IS_WARNING(code)) warningCount++;
    }

    if(specialCase < 1)
    {
        lexan.reportCurrentFileName(errorStream);

        if(SE_IS_ERROR(code)) fprintf(errorStream, "ERROR");
        else if(SE_IS_WARNING(code)) fprintf(errorStream, "WARNING");

        fprintf(errorStream, "(line %d): %s\n",
            lexan.getLineNum(isParserLookahead()), ErrStr);
    }

        // If there are too many semantic errors, the input will be
        //  interrupted. We need to inform the user about it:
    if(semanticErrorCount == MAX_SEMANTIC_ERRORS)
    {
        fprintf(errorStream, "TOO MANY SEMANTIC ERRORS; parsing is stopped; "
            "no output will be produced\n");
    }

        // Interrupt the input if too many errors
    if(semanticErrorCount >= MAX_SEMANTIC_ERRORS) lexan.interruptInput();

        // End the arguments and return:
    va_end(args);
    return 0;
}


	// This method reports the program variables
	// that have been declared but never used
void Rules::reportUnusedVariables()
{
	VarNode *curVar;
	int isPrinted;

		// Check declared and used variables:
	for(curVar=first.var,isPrinted=0;curVar;curVar=curVar->next)
	{
		if(curVar->valIndex==VAR_NOTUSED)
		{  // Found the variable that was declared but never used
			// If this is the first time, print the message
			if(!isPrinted)
			  fprintf(errorStream, "WARNING: The following program "
			    "variables were declared but never used:  ");
			else
				fprintf(errorStream, ", ");

			// Print the variable itself:
			fprintf(errorStream, "%s",
				symbt.getName(curVar->symbolno));
			isPrinted=1;
		}
	}

		// If the message was printed, print the newline and
		//  count this as a warning
	if(isPrinted)
	{
		fprintf(errorStream, "\n");
		warningCount++;
	}
}


	// This method reports the program predicates
	// that have been declared but never used
void Rules::reportUnusedProgramPredicates()
{
	PPnode *pp;
	int isFirst;

		// Check for predicates that were declared but never used:
		//  Very similar to the method 'reportUnusedVariables'.
	for(pp=first.pred, isFirst=1 ; pp ; pp=pp->next)
		if(pp->usage==PP_DECLARED)
		{
			if(isFirst)
			   fprintf(errorStream, "WARNING:The following program "
			   "predicates were declared but never used (they WILL "
			   "NOT appear in the ground theory): ");
			else
				fprintf(errorStream, ", ");
			isFirst=0;
			fprintf(errorStream, symbt.getName(pp->symbolno));
		}

	if(!isFirst)
	{
		fprintf(errorStream, "\n");
		warningCount++;
	}
}


	// This method is called by the function 'error' whenever syntax error
	// occurrs. It increases syntax error count and if there are too much
	// errors, it interrupts the parsing and informs the user about this
void Rules::registerSyntaxError()
{
	syntaxErrorCount++;

	if(syntaxErrorCount > MAX_SYNTAX_ERRORS)
	{
		fprintf(stderr, "TOO MANY SYNTAX ERRORS; parsing is stopped; "
			"no output will be produced\n");
		lexan.interruptInput();
	}
}




/****************************************************************************
*****************************************************************************
******  METHODS FOR PRINTING DEBUG INFORMATION in the class 'Rules'
*****************************************************************************
****************************************************************************/

// Auxiliary constants for printing:
#define PRINT_NUMS_PER_LINE 10
#define PRINT_WORDS_PER_LINE 15

// NOTE: All the debugging methods print the information to the standard output

	// Central debugging function. Prints a lot of useful information
	// (although not everything).
void Rules::test()
{
	printf("Printing debug information: \n");

		// Print string constants, data predicates, weight functions
		// program variables and program predicates.
	printStringConstants();

	printDataPredicates();
	printWeights();

	printVariables();
	printProgramPredicates();

		// Print memory usage information:
	#ifdef MEMORY_H_DEBUG
	printf("\nPermanent memory source information: \n");
	permMSrc.printStatistics();
	printf("\nShort-term temporary node source information: \n");
	tempSrc.printStatistics();
	printf("\nValue node source information: \n");
	valSrc.printStatistics();
	printf("\nVariable node source information: \n");
	varSrc.printStatistics();
	printf("\nExpression node source information: \n");
	expSrc.printStatistics();
	printf("\nQuantifier node source information: \n");
	quantSrc.printStatistics();
	printf("\nAtom node source information: \n");
	atomSrc.printStatistics();
	printf("\nMeta Atom node source information: \n");
	metaSrc.printStatistics();
	printf("\nRule node source information: \n");
	ruleSrc.printStatistics();
	#endif
}


	// This method prints the whole theory using 'printRule' method
void Rules::printTheory()
{
	for(RuleNode *rule=getRules(); rule;rule=rule->next)
		printRule(rule);
}

	// This method prints a permanent rule:
void Rules::printRule(RuleNode *rule)
{
	printf("%d: ", rule->lineNum);  // Print line number
		// Print quantifier and conditions for the rule:
	printf("[");
	printQuantifierList(rule->quant);
	printf("]: ");

		// Print the head if this is a horn rule:
	if(rule->horn)
	{
		printProgramAtomList(rule->head, "");
		printf(":- ");
	}

		// Print the body of the rule:
		// Note: put ',' between atoms:
	printProgramAtomList(rule->body, ",");
	if(rule->body && rule->cardBody) printf(",");
	printMetaAtomList(rule->cardBody, ",");
	if(rule->weightBody && (rule->cardBody || rule->body)) printf(",");
	printMetaAtomList(rule->weightBody, ",");

		// Print the head if this is not a horn rule:
	if(!rule->horn)
	{
		printf(" -> ");
			// Note: put '|' between atoms:
		printProgramAtomList(rule->head, "|");
		if(rule->head && rule->cardHead) printf("|");
		printMetaAtomList(rule->cardHead, "|");
		if(rule->weightHead && (rule->cardHead || rule->head))
			printf("|");
		printMetaAtomList(rule->weightHead, "|");
	}

	printf("\n");
}


/*
void Rules::printTempRule(TempAtom *antecedent, TempAtom *consequent, int horn)
DESCRIPTION
	Prints temporary rule
PARAMETERS
	antecedent   -  the body of the rule
	consequent   -  the head of the rule
	horn         -  TRUE for horn rules, FALSE for implications
*/
void Rules::printTempRule(TempAtom *antecedent, TempAtom *consequent, int horn)
{
		// List pointed by 'left' will be printed first.
	TempAtom *left, *right;

		// Determine what to print first and what last: head or body.
	if(horn) {left=consequent; right=antecedent;}
	else {left=antecedent; right=consequent;}

		// Print the first list:
	for(;left;left=left->next)
	{
		printTempAtom(left);
		if(left->next) printf(" ");
	}

		// Print separator between the head and the body:
	if(horn) printf(" :- "); else printf(" -> ");

		// Print the second list:
	for(;right;right=right->next)
	{
		printTempAtom(right);
		if(right->next) printf(" ");
	}

	printf("\n");
}


	// This method prints the list of quantifiers that can also
	//  contain conditions:
void Rules::printQuantifierList(Quantifier *list)
{
	for(; list ; list=list->next)
	{
			// If this is not an empty quantifier, print the
			//  objects that are quantified:
		if(list->token!=QUANT_EMPTY)
		{
			// If variable quantifier, print the variable:
			if(list->token==QUANT_VAR)
				PRINT_VARIABLE(list->var)
			else  // If data predicate quantifier, print
			      // the data predicate with its arguments:
				printObjectWithArgs(
					symbt.getName(list->dp->symbolno),
					list->dp->arity,list->arg, 0);
		}
			// Print the list of conditions if present:
		if(list->conditions)
		{
			printf(":(");
			printExpression(list->conditions);
			printf(")");
		}

		if(list->next) printf(","); // Print the separator
	}
}


/*
void Rules::printMetaAtomList(MetaAtom *list, char *separator)
DESCRIPTION
	This method can be used to print permanent list of meta atoms
	in both body and head.
PARAMETERS
	list      - List of meta atoms to print
	separator - String that will be printed between the atoms.
	             It should be "," for body and "|" for head.
*/
void Rules::printMetaAtomList(MetaAtom *list, char *separator)
{
	for(; list ; list=list->next)
	{
		printExpression(list->min);   // Print MIN
		printf("{");
			// Print atom list separated by ";".
		printProgramAtomList(list->list, ";");
		printf("}");
		printExpression(list->max);  // Print MAX
		if(list->next) printf(separator);  // Print separator
	}
}


/*
void Rules::printProgramAtomList(Atom *list, char *separator)
DESCRIPTION
	This method can be used to print permanent list of program atoms
	in body, head, or inside meta-atoms.
PARAMETERS
	list      - List of program atoms to print
	separator - String that will be printed between the atoms.
	             It should be "," for body,  "|" for head
	             and ";" for meta-atom.
*/
void Rules::printProgramAtomList(Atom *list, char *separator)
{
	PPnode *pp;

	for( ; list ; list=list->next)
	{
			// Print the program predicate and the arguments:
		pp=list->pp;
		printObjectWithArgs(symbt.getName(pp->symbolno),
				pp->arity,list->arg, pp->params);

			// If there is a weight expression, print it:
		if(list->weight)
		{
			printf("=\"");
			printExpression(list->weight);
			printf("\"");
		}

			// If there are local quantifiers, print them:
		if(list->quant)
		{
			printf("[");
			printQuantifierList(list->quant);
			printf("]");
		}

		if(list->next) printf(separator); // Print the separator
	}
}


	// This method prints a temporary atom
void Rules::printTempAtom(TempAtom *atom)
{
	PPnode *pp;
	DPnode *dp;
	TempAtom *list;

	if(!atom) return;

	switch(atom->type)
	{
	case ATOM_PROG:  // If this s a program atom:
		    // Print the program predicate with arguments:
		pp=atom->prog.pp;
		printObjectWithArgs(symbt.getName(pp->symbolno),
			pp->arity,atom->prog.arg, pp->params);
		    // Print weight expression:
		if(atom->prog.weight)
		{
			printf("=\"");
			printExpression(atom->prog.weight);
			printf("\"");
		}

		    // Print local variables or quantifiers:
		if(atom->prog.quant)
		{
			printf("[");
			printQuantifierList(atom->prog.quant);
			printf("]");
		}

		     // Print the data restrictions:
		for(list=atom->prog.restrictions;list;list=list->next)
		{
			printf(":");
			printTempAtom(list);
		}
		break;
	case ATOM_WEIGHT:
	case ATOM_CARD:   // For cardinality or weight atom:
		printExpression(atom->meta.min);  // Prrint MIN
		printf("{");
			// Print meta-list:
		for(list=atom->meta.list;list;list=list->next)
		{
			printTempAtom(list);
			printf(" ");
		}
		printf("}");
		printExpression(atom->meta.max); // Print MAX
		break;
	case ATOM_CONDITION: // For predefined atom: print the expression
		printf("\"");
		printExpression(atom->cond.first);
		printf("\"");
		break;
	case ATOM_DATANEG:
		printf("~");  // Put this for negative data predicate condition
	case ATOM_DATA:
		    // Print the data predicate and arguments:
		dp=atom->data.dp;
		printObjectWithArgs(symbt.getName(dp->symbolno),
			dp->arity,atom->data.arg, 0);
		break;
	    // Print preliminary true or false atoms:
	case ATOM_TRUE:  printf(" !TRUE! ");  break;
	case ATOM_FALSE: printf(" !FALSE! "); break;
	default:
		fprintf(stderr, "CRITICAL ERROR while "
			"printing of temporary atom: \n"
			"Unknown type of atom given: %d\n\n",atom->type);
		exit(1);
	}
}



/*
void Rules::printObjectWithArgs(char *name, int arity,
                                Expression *args, DPnode **types)
DESCRIPTION
	Universal method can print any object with arguments:
	data atom, program atom, or weight function call.
PARAMETERS
	name   -  Name of the object (usually from the symbol table)
	arity  -  Arity of the object
	args   -  Argument list of the object
	types  -  Array of pointers to the DPnode that correspond to
	           a type of each argument. Necessary only if some
	           arguments can have type EXPR_INDEX (only for
	           progaram atoms).
*/
void Rules::printObjectWithArgs(char *name, int arity,
                                Expression *args, DPnode **types)
{
	int i;
	Expression *cur;

	printf("%s", name);  // Print the name of the object:
	if(arity>0)  // Print the arguments if necessary
	{
		printf("(");

			// Go over the argument list:
		for(i=0, cur=args ; i<arity && cur ; i++, cur=cur->next)
		{
				// Print the argument depending on its token:
			switch(cur->token)
			{
			case EXPR_STRING:
				strConst.printConstant(stdout, cur->num);
				break;
			case EXPR_NUM:
				printf("%d", cur->num);
				break;
			case EXPR_INDEX:
				// In this case we have to have 'types' array:
				if(!types)
				{
				  fprintf(stderr,"CRITICAL ERROR while "
				     "printing an object with arguments.\n"
				     "The object must have a type in order"
				     "to have argument EXPR_INDEX.\n\n");
				  exit(1);
				}
				// Print the constant taken from the argument
				// list of the data type of the current argument
				strConst.printConstant(stdout,
				         types[i]->arguments[cur->num]);
				break;
			case EXPR_VAR_QUANT:
				// Use this to mark quantified variables:
				printf("#");
			case EXPR_VAR:
				PRINT_VARIABLE(cur->var);
				break;
			case EXPR_EXPR:
				printExpression(cur->exp);
				break;
			case EXPR_UNDERSCORE:
				printf("_");
				break;
			default:
				fprintf(stderr,"CRITICAL ERROR while printing"
				  " an object with arguments.\n"
				  "Unknown argument type: %d\n\n", cur->token);
				exit(1);
			}
				// Print the separator between the arguments:
			if(i<arity-1) printf(",");
		}
		printf(")");
	}
}


	// This method prints an arithmetic expression:
void Rules::printExpression(Expression *exp)
{
	char *str;

	for(; exp ; exp=exp->next)
	{
		str=0;

		switch(exp->token)
		{
		// Print constant, number, or variable:
		case EXPR_STRING:
			strConst.printConstant(stdout,exp->num);
			break;
		case EXPR_NUM:
			printf("%d",exp->num);
			break;
		case EXPR_VAR:
			PRINT_VARIABLE(exp->var);
			break;
		case EXPR_EXPR:
			printf("(");
			printExpression(exp->exp);
			printf(")");
			break;

		// Print relational operator:
		case EXPR_LESS:       str="<" ; break;
		case EXPR_NOTGREATER: str="<="; break;
		case EXPR_GREATER:    str=">" ; break;
		case EXPR_NOTLESS:    str=">="; break;
		case EXPR_EQUAL:      str="=="; break;
		case EXPR_NOTEQUAL:   str="!="; break;

		// Print arithmetic operator:
		case EXPR_PLUS:  str="+"; break;
		case EXPR_MINUS: str="-"; break;
		case EXPR_MUL:   str="*"; break;
		case EXPR_DIV:   str="/"; break;

		// Print binary function:
		case EXPR_MOD:    str="[MOD]"; break;
		case EXPR_MIN:    str="[MIN]"; break;
		case EXPR_MAX:    str="[MAX]"; break;

		// Print unary function
		case EXPR_ABS:	  str="[ABS]"; break;
		case EXPR_UMINUS: str="[NEG]"; break;

		// Print weight function:
		case EXPR_WEIGHT:
			printf("%s ",symbt.getName(exp->wFun->symbolno));
			break;

		// Print positive or negative data atom:
		case EXPR_DNEGATIVE:
			printf("~");
		case EXPR_DPOSITIVE:
			// Print data atom:
			printObjectWithArgs(
				symbt.getName(exp->atom->pp->symbolno),
				exp->atom->pp->arity, exp->atom->arg,
				exp->atom->pp->params);
			break;

		// Print assignment operation:
		case EXPR_ASSIGN:
			printf("=(");
			PRINT_VARIABLE(exp->var);
			printf(")");
			break;

		default:
			printf("\nERROR: Unknown expression "
				"with the token: %d\n", exp->token);
		}

		if(str) printf("%s",str); // Print in the simple case

		if(exp->next) printf(" "); // Print separator
	}
}


	// This method prints all the string constants in the program
	//  It can be called before the data sequence is finished as well
	//  as after.
void Rules::printStringConstants()
{
	int i, min, max, num;

	printf("Printing all string constants:\n");

		// First, if data sequence is finished, check for consistency;
		//  otherwise, print.
	symbt.getNumRange(&min, &max);
	for(i=min; i<=max;i++)
	{
		if(symbt.getType(i) != TYPE_STRING_CONST) continue;

		if(isDataFinished)
		{
			num=strConst.getNumberForConstant(i);

				// Check consistency of building the constants:
			if(strConst.getConstantFromNumber(num) !=
				symbt.getName(i))
			{
				fprintf(stderr, "CRITICAL ERROR in "
					"Rules::printStringConstants\n"
					"Detected logical error in "
					"class \'StringConstants\'\n\n");
				exit(1);
			}
		}
		else
			printf("%s, ", symbt.getName(i));
	}

		// Cannot do anything else if data sequence is not finished.
	if(!isDataFinished) return;

		// Get the minimum and the maximum numbers that are treated
		//  as string constants:
	strConst.getConstantRange(&min, &max);

		// Print all the constants:
	for(i=min; i<=max;i++)
		printf("%s(%d), ", strConst.getConstantFromNumber(i), i);

	printf("\n\n");
}


	// This method prints all the data predicates:
void Rules::printDataPredicates()
{
	int i, i1;
	DPnode *data;
	Value *val;
	int sno, count, vCount, LineStart;
	printf("Printing all data predicates: \n");

		// Go over the list of data predicates and weight functions:
	for(count=0, data=first.data; data ; data=data->next)
	{
			// Check consistency in the symbol table:
		if(symbt.getVal(data->symbolno)->ptr != (void *)data)
		{
			fprintf(stderr, "CRITICAL ERROR detected in "
				"Rules::printDataPredicates\n\n");
			exit(1);
		}

			// Not a data predicate => go to the next element:
		if(symbt.getType(data->symbolno) != TYPE_DATA_PRED) continue;

		count++;
		sno=data->symbolno;
			// Print the basic information about the predicate:
		printf("%d)[ %s ] \n", count, symbt.getName(sno));
		printf("  symbol number:\t%d\n", data->symbolno);
		printf("  arity:\t\t%d\n", data->arity);
		printf("  number of values:\t%d\n", data->valNum);
		printf("  values: ");

		if(isDataFinished)
		{
				// If data sequence is finished, the arguments
				// are stored in the array or vectors
			vCount=1;
			LineStart=0;

				// For every vector in the array of arguments:
			for(i=0 ; i<data->valNum*data->arity ; i+=data->arity)
			{
					// Print all the elements of the vector:
				for(i1=0 ; i1<data->arity ; i1++, vCount++)
				{
					strConst.printConstant(stdout,
						data->arguments[i+i1]);
					printf(" ");
				}

				printf("| "); // print the separator

					// Go to the next line if necessary
				if(vCount>LineStart+PRINT_NUMS_PER_LINE)
				{
					printf("\n          ");
					LineStart=vCount;
				}
			}
			printf("\n\n");
		}
		else
		{
				// If we are still in the data sequence,
				//  the arguments are lists of 'Value' nodes:
			vCount=1;
			LineStart=0;
			for(val=data->tmpArguments;val;val=val->next,vCount++)
			{
				// Print the argument: if it is a number, print
				// the number in 'val->num'; otherwise, print
				// the name of the constant with the symbol
				// number 'val->num'.
				if(val->type==EXPR_NUM)
				      printf("%3d(%d) ",val->num,val->lineNum);
				else
				      printf("%s(%d) ",
				         symbt.getName(val->num),val->lineNum);

				// Check whether we are between two vectors:
				if(!(vCount%data->arity)) // If so:
				{
					printf("| ");  // Print the separator
					// Check whether we need a new line
					if(vCount>LineStart+PRINT_NUMS_PER_LINE)
					{
						printf("\n          ");
						LineStart=vCount;
					}
				}
			}
			printf("\n\n");
		}
	}
	printf("\n");
}


	// This method prints all the weight functions:
void Rules::printWeights()
{
	WFnode *wn;
	int count;
	int vCount, LineStart;
	int i, i1;

	printf("Printing all weight functions:\n");
	
	for(count=0,wn=first.data; wn ;wn=wn->next)
	{
			// Check consistency in the symbol table:
		if(symbt.getVal(wn->symbolno)->ptr != (void *)wn)
		{
			fprintf(stderr, "CRITICAL ERROR detected in "
				"Rules::printWeights\n\n");
			exit(1);
		}

		if(symbt.getType(wn->symbolno)!=TYPE_WEIGHT_FUN) continue;

		count++;
			// Print the basic information:
		printf("%d)[ %s ] \n", count, symbt.getName(wn->symbolno));
		printf("  symbol number:\t%d\n", wn->symbolno);
		printf("  arity:\t\t%d\n", wn->arity);
		printf("  number of values:\t%d\n", wn->valNum);
		printf("  Function: ");

			// This method prints the information only if the
			//  declaration sequence is finished
		if(isDataFinished)
		{
			vCount=1;
			LineStart=0;
			for(i=0 ; i<wn->valNum*wn->arity ; i+=wn->arity)
			{
				printf("(");
					// Print the first argument:
				strConst.printConstant(stdout,wn->arguments[i]);

					// Print the rest of the arguments:
				for(i1=1;i1<wn->arity;i1++,vCount++)
				{
					printf(",");
					strConst.printConstant(stdout,
						wn->arguments[i+i1]);
				}

					// Print the result
				printf(")=%d  ",wn->results[i/wn->arity]);

					// Print the newline if necessary:
				if(vCount>LineStart+PRINT_NUMS_PER_LINE)
				{
					printf("\n          ");
					LineStart=vCount;
				}
			}
			printf("\n\n");
		}
		else
			printf("Cannot print the weight function values\n"
				"before the lists of arguments "
				"have been converted into arrays\n\n");
	}
}


	// This method prints all the program variables:
void Rules::printVariables()
{
	VarNode *var;
	char *name, *oldName;
	DPnode *type;

	printf("Printing all program variables: \n");

	for(var=first.var,oldName=0;var;var=var->next)
	{
		type=var->type;

			// Get the name of the variable type:
		if(type==&DPError) name="ERROR IN TYPE";
		else name=symbt.getName(type->symbolno);

			// If the type of this variable is the same as of the
			//  previous one, print the separator.
			// Otherwise, print the "var nameOfDataType"
		if(name != oldName)
		{
			if(oldName) printf("\n");
			printf("var %s ",name);
		}
		else
			printf(", ");

		oldName=name;
			// Print the name of the variable:
		printf("%s", symbt.getName(var->symbolno));
	}
	printf("\n\n");
}


	// This method prints all the program predicates:
void Rules::printProgramPredicates()
{
	PPnode *pp;
	char *name;
	DPnode *type;
	int count;
	int words;
	int i;

	printf("Printing all program predicates:\n"
	  "Meaning of symbols: ~ - horn predicate, ^ - meta atom predicate, \n"
	  "\t\t # - error predicate, * - not used in any rule\n");

		// Go over the list of program predicates:
	for(count=PRINT_WORDS_PER_LINE, pp=first.pred ; pp ; pp=pp->next)
	{
			// Check whether we need to go to the new line:
		words=pp->arity + 1 + ((pp->tail)?1:0);
		if(words+count<=PRINT_WORDS_PER_LINE) count+=words;
		else {printf("\n"); count=words;}

			// Print the symbol before the predicate:
		if(pp->usage == PP_DECLARED) printf("*");
		switch(pp->hornMeta)
		{
		case PP_HORN:  printf("~"); break;
		case PP_META:  printf("^"); break;
		case PP_WASTED:printf("#"); break;
		}

			// Print the predicate name:
		printf("%s",symbt.getName(pp->symbolno));

			// Print the types of arguments if necessary:
		if(pp->arity>0)
		{
			printf("(");

				// Print the types:
			for(i=0 ; i < pp->arity ; i++)
			{
					// Get the name of the type:
				type=pp->params[i];
				if(type==&DPError) name="ERROR IN TYPE";
				else name=symbt.getName(type->symbolno);

				if(i>0) printf(","); // Print the separator
				printf("%s",name);   // Print the name
			}

				// Print the tail if necessary:
			if(pp->tail) printf("):%s, ",
				symbt.getName(pp->tail->symbolno));
			else printf("), ");
		}
		else
			printf(", ");
	}
	printf("\n\n");
}
